//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 */
//  ========================================================================

/**
 * @type {TP.sherpa.domvisualizer}
 */

//  ------------------------------------------------------------------------

TP.sherpa.TemplatedTag.defineSubtype('domvisualizer');

//  ------------------------------------------------------------------------
//  Type Constants
//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Type.defineConstant('SORTABLE_LOCATION',
            'https://cdn.jsdelivr.net/npm/sortablejs@latest/Sortable.min.js');

//  ------------------------------------------------------------------------
//  Type Methods
//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Type.defineMethod('tagAttachDOM',
function(aRequest) {

    /**
     * @method tagAttachDOM
     * @summary Sets up runtime machinery for the element in aRequest
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,
        scriptLoc;

    //  this makes sure we maintain parent processing
    this.callNextMethod();

    //  Make sure that we have an Element to work from
    if (!TP.isElement(elem = aRequest.at('node'))) {
        //  TODO: Raise an exception.
        return;
    }

    //  If ReactDOM isn't loaded, then we must load it *into the target document
    //  of the element that we're processing*.
    if (TP.notValid(TP.nodeGetWindow(elem).Sortable)) {
        scriptLoc = this.SORTABLE_LOCATION;
        TP.sys.fetchScriptInto(
                TP.uc(scriptLoc),
                TP.nodeGetDocument(elem),
                null,
                TP.hc('crossorigin', true)).then(
                    function() {
                        TP.wrap(elem).setup();
                    });
    } else {
        TP.wrap(elem).setup();
    }

    return;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Type.defineMethod('tagDetachDOM',
function(aRequest) {

    /**
     * @method tagDetachDOM
     * @summary Tears down runtime machinery for the element in aRequest.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem;

    //  Make sure that we have an Element to work from
    if (!TP.isElement(elem = aRequest.at('node'))) {
        //  TODO: Raise an exception.
        return;
    }

    TP.wrap(elem).teardown();

    //  this makes sure we maintain parent processing - but we need to do it
    //  last because it nulls out our wrapper reference.
    this.callNextMethod();

    return;
});

//  ------------------------------------------------------------------------
//  Instance Attributes
//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineAttribute('$lastContextMenuSignal');

TP.sherpa.domvisualizer.Inst.defineAttribute('listcontent',
    TP.cpc('> .content', TP.hc('shouldCollapse', true)));

TP.sherpa.domvisualizer.Inst.defineAttribute('listitems',
    TP.cpc('> .content > li:not(.spacer)', TP.hc('shouldCollapse', false)));

//  ------------------------------------------------------------------------
//  Instance Methods
//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineMethod('focusOnTarget',
function(aTPElement) {

    /**
     * @method focusOnTarget
     * @summary Focuses the receiver onto the supplied target.
     * @param {TP.dom.UIElementNode} aTPElement The element to focus the
     *     receiver on.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var targetElem,
        targetDoc,
        thisWin,

        str,

        treeWalker,
        traverseLevel,

        newContent,
        newElem,
        nestedSortables,

        len,
        i;

    targetElem = aTPElement.getNativeNode();
    targetDoc = aTPElement.getNativeDocument();

    thisWin = this.getNativeWindow();

    str = '<div class="list-group col nested-sortable">';

    treeWalker = targetDoc.createTreeWalker(
                                targetElem, NodeFilter.SHOW_ELEMENT);

    traverseLevel = function(walker, indent) {

        //  describe current node
        var parent,
            child;

        parent = walker.currentNode;

        str += '<div class="list-group-item nested-' + indent + '">' +
                '<div class="list-group-label">' + parent.tagName + '</div>';

        // traverse children:
        for (child = walker.firstChild();
                child !== null;
                    child = walker.nextSibling()) {

            str += '<div class="list-group nested-sortable">';
            traverseLevel(walker, indent + 1);
            str += '</div>';
        }

        str += '</div>';

        //  return position to the current (level up):
        walker.currentNode = parent;
    }

    traverseLevel(treeWalker, 1);

    str += '</div>';

    newContent = this.setContent(TP.xhtmlnode(str));

    newElem = newContent.getNativeNode();

    nestedSortables = TP.byCSSPath('.nested-sortable', newElem, false, false);

    len = nestedSortables.getSize();
    for (i = 0; i < len; i++) {

        /* eslint-disable no-new */
        new thisWin.Sortable(
                nestedSortables.at(i),
                {
                    group: 'nested',
                    animation: 150,
                    fallbackOnBody: true,
                    swapThreshold: 0.65
                });
        /* eslint-enable no-new */
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineMethod('getContentTagForContextMenu',
function(aSignal) {

    /**
     * @method getContentTagForContextMenu
     * @summary Returns the tag name of the content to use in a context menu.
     *     Note that this should return the plain (possibly namespaced) name
     *     with no markup bracketing, etc.
     * @param {TP.sig.ShowContextMenu} aSignal The TIBET signal which triggered
     *     the context menu to show and menu content to be required.
     * @returns {String} The name of the tag to use as content for the context
     *     menu.
     */

    return null;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('CanvasChanged',
function(aSignal) {

    /**
     * @method handleCanvasChanged
     * @summary Handles notifications of the canvas changing from the Sherpa
     *     object.
     * @param {TP.sig.CanvasChanged} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var halo,
        haloTarget;

    //  Grab the halo and make sure it's focused. If not, we just bail out here.
    halo = TP.byId('SherpaHalo', this.getNativeDocument());
    if (!halo.isFocused()) {
        return this;
    }

    haloTarget = halo.get('currentTargetTPElem');

    this.focusOnTarget(haloTarget);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('DocumentLoaded',
function(aSignal) {

    /**
     * @method handleDocumentLoaded
     * @summary Handles when the document in the current UI canvas loads.
     * @param {TP.sig.DocumentLoaded} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('DocumentUnloaded',
function(aSignal) {

    /**
     * @method handleDocumentUnloaded
     * @summary Handles when the document in the current UI canvas unloads.
     * @param {TP.sig.DocumentUnloaded} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('PclassClosedChange',
function(aSignal) {

    /**
     * @method handlePclassClosedChangeFromSherpaHUD
     * @summary Handles notifications of HUD closed change signals.
     * @param {TP.sig.PclassClosedChange} aSignal The TIBET signal which
     *     triggered this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var hudIsClosed;

    //  Grab the HUD and see if it's currently open or closed.
    hudIsClosed = TP.bc(aSignal.getOrigin().getAttribute('closed'));

    if (hudIsClosed) {
        this.toggleObservations(false);
    } else {
        this.toggleObservations(true);
    }

    return this;
}, {
    origin: 'SherpaHUD'
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('HaloDidBlur',
function(aSignal) {

    /**
     * @method handleHaloDidBlur
     * @summary Handles notifications of when the halo blurs on an object.
     * @param {TP.sig.HaloDidBlur} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var sherpaInst;

    this.setValue(null);

    sherpaInst = TP.bySystemId('Sherpa');
    this.ignore(sherpaInst, 'CanvasChanged');

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('HaloDidFocus',
function(aSignal) {

    /**
     * @method handleHaloDidFocus
     * @summary Handles notifications of when the halo focuses on an object.
     * @param {TP.sig.HaloDidFocus} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var haloTarget,

        sherpaInst;

    haloTarget = aSignal.at('haloTarget');

    this.focusOnTarget(haloTarget);

    sherpaInst = TP.bySystemId('Sherpa');
    this.observe(sherpaInst, 'CanvasChanged');

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('ScreenWillToggle',
function(aSignal) {

    /**
     * @method handleScreenWillToggle
     * @summary Handles notifications of screen will toggle signals.
     * @param {TP.sig.ScreenWillToggle} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var world,

        oldScreenTPWin;

    world = TP.byId('SherpaWorld', TP.sys.getUIRoot());

    //  Grab the old screen TP.core.Window and ignore
    //  DocumentLoaded/DocumentUnloaded signals coming from it.
    oldScreenTPWin = world.get('selectedScreen').getContentWindow();
    this.ignore(oldScreenTPWin, TP.ac('DocumentLoaded', 'DocumentUnloaded'));

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineHandler('ScreenDidToggle',
function(aSignal) {

    /**
     * @method handleScreenDidToggle
     * @summary Handles notifications of screen did toggle signals.
     * @param {TP.sig.ScreenDidToggle} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var world,

        newScreen,
        newScreenTPWin;

    world = TP.byId('SherpaWorld', TP.sys.getUIRoot());

    //  Grab the new screen TP.core.Window and observe
    //  DocumentLoaded/DocumentUnloaded signals coming from it.
    newScreen = world.get('screens').at(aSignal.at('screenIndex'));

    if (TP.isValid(newScreen)) {
        newScreenTPWin = newScreen.getContentWindow();
        this.observe(newScreenTPWin,
                        TP.ac('DocumentLoaded', 'DocumentUnloaded'));
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineMethod('setup',
function() {

    /**
     * @method setup
     * @summary Perform the initial setup for the receiver.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var haloTPElem,
        haloTargetTPElem;

    this.observe(TP.byId('SherpaHalo', this.getNativeDocument()),
                    TP.ac('TP.sig.HaloDidFocus', 'TP.sig.HaloDidBlur'));

    this.observe(TP.byId('SherpaHUD', this.getNativeDocument()),
                    'PclassClosedChange');

    //  The target element that we're moving is the halo's current target.
    haloTPElem = TP.byId('SherpaHalo', TP.sys.getUIRoot());
    haloTargetTPElem = haloTPElem.get('currentTargetTPElem');

    this.focusOnTarget(haloTargetTPElem);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineMethod('teardown',
function() {

    /**
     * @method teardown
     * @summary Tears down the receiver by performing housekeeping cleanup, like
     *     ignoring signals it's observing, etc.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    this.ignore(TP.byId('SherpaHalo', this.getNativeDocument()),
                TP.ac('TP.sig.HaloDidFocus', 'TP.sig.HaloDidBlur'));

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.domvisualizer.Inst.defineMethod('toggleObservations',
function(shouldObserve) {

    /**
     * @method toggleObservations
     * @summary Either observe or ignore the signals that the receiver needs to
     *     function.
     * @param {Boolean} shouldObserve Whether or not we should be observing (or
     *     ignoring) signals.
     * @returns {TP.sherpa.domvisualizer} The receiver.
     */

    var world,
        currentScreenTPWin;

    world = TP.byId('SherpaWorld', TP.sys.getUIRoot());
    currentScreenTPWin = world.get('selectedScreen').getContentWindow();

    if (shouldObserve) {
        this.observe(world, TP.ac('ScreenWillToggle', 'ScreenDidToggle'));
        this.observe(currentScreenTPWin,
                        TP.ac('DocumentLoaded', 'DocumentUnloaded'));
    } else {
        this.ignore(world, TP.ac('ScreenWillToggle', 'ScreenDidToggle'));
        this.ignore(currentScreenTPWin,
                        TP.ac('DocumentLoaded', 'DocumentUnloaded'));
    }

    return this;
});

//  ------------------------------------------------------------------------
//  end
//  ========================================================================
