//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 */
//  ========================================================================

/**
 * @type {TP.sherpa.dimensionsManipulator}
 */

//  ------------------------------------------------------------------------

TP.sherpa.TemplatedTag.defineSubtype('dimensionsManipulator');

//  ------------------------------------------------------------------------
//  Type Methods
//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Type.defineMethod('tagAttachDOM',
function(aRequest) {

    /**
     * @method tagAttachDOM
     * @summary Sets up runtime machinery for the element in aRequest
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,
        tpElem;

    //  this makes sure we maintain parent processing
    this.callNextMethod();

    //  Make sure that we have an Element to work from
    if (!TP.isElement(elem = aRequest.at('node'))) {
        //  TODO: Raise an exception.
        return;
    }

    tpElem = TP.wrap(elem);

    tpElem.setup();

    return;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Type.defineMethod('tagDetachDOM',
function(aRequest) {

    /**
     * @method tagDetachDOM
     * @summary Tears down runtime machinery for the element in aRequest.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,
        tpElem;

    //  Make sure that we have an Element to work from
    if (!TP.isElement(elem = aRequest.at('node'))) {
        //  TODO: Raise an exception.
        return;
    }

    tpElem = TP.wrap(elem);

    tpElem.ignore(TP.byId('SherpaHalo', TP.sys.getUIRoot()),
                    'TP.sig.HaloDidFocus');

    tpElem.ignore(TP.byId('SherpaHalo', TP.sys.getUIRoot()),
                    'TP.sig.HaloDidBlur');

    //  this makes sure we maintain parent processing - but we need to do it
    //  last because it nulls out our wrapper reference.
    this.callNextMethod();

    return;
});

//  ------------------------------------------------------------------------
//  Instance Attributes
//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineAttribute('$currentModifyingRule');

TP.sherpa.dimensionsManipulator.Inst.defineAttribute(
    'northguide',
    TP.cpc('> #northguide', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsManipulator.Inst.defineAttribute(
    'westguide',
    TP.cpc('> #westguide', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsManipulator.Inst.defineAttribute(
    'southguide',
    TP.cpc('> #southguide', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsManipulator.Inst.defineAttribute(
    'eastguide',
    TP.cpc('> #eastguide', TP.hc('shouldCollapse', true)));

//  ------------------------------------------------------------------------
//  Instance Methods
//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineMethod('activate',
function(aTargetTPElem, aSignal) {

    /**
     * @method activate
     * @summary Activates the receiver.
     * @param {TP.dom.ElementNode} aTargetTPElem The element that the receiver
     *     will be activated on.
     * @param {TP.sig.Signal} [aSignal] An optional signal that caused the
     *     receiver to activate.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    //  Check to make sure we can adjust the 'width' property on
    //  the target element (assuming we can adjust 'height' as
    //  well).
    var cantAdjustReasons;

    cantAdjustReasons = aTargetTPElem.canAdjustCSSProperty('width');

    if (TP.notEmpty(cantAdjustReasons)) {
        if (cantAdjustReasons.contains(TP.ELEMENT_NEEDS_TO_BE_BLOCK)) {
            TP.prompt(
                'This element needs a "display" property value of:' +
                ' "block", "inline-block", "list-item", "run-in", "table" or' +
                ' "table-cell" to be sizable. Enter a value for the "display"' +
                ' property:', 'block').then(
                function(displayVal) {
                    var stylesHUD,
                        modifyingRule;

                    if (TP.isEmpty(displayVal)) {
                        return;
                    }

                    if (displayVal === 'block' ||
                        displayVal === 'inline-block' ||
                        displayVal === 'list-item' ||
                        displayVal === 'run-in' ||
                        displayVal === 'table' ||
                        displayVal === 'table-cell') {

                        stylesHUD = TP.byId('StylesHUD', TP.win('UIROOT'));
                        modifyingRule = stylesHUD.getModifiableRule();

                        if (TP.notValid(modifyingRule)) {
                            return this;
                        }

                        TP.styleRuleSetProperty(
                                modifyingRule, 'display', displayVal, true);

                    } else {
                        TP.alert(
                            'Invalid "display" property value: ' + displayVal);
                    }
                });
            }

        //  Return here because we don't want to continue with activation.
        return this;
    }

    this.setAttribute('hidden', false);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineMethod('resizeToTarget',
function(aTPElement) {

    /**
     * @method resizeToTarget
     * @summary Focuses the receiver onto the supplied target.
     * @param {TP.dom.UIElementNode} aTPElement The element to focus the
     *     receiver on.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    var halo,

        ourRect,
        targetRect,

        x,
        y,

        northRect,
        westRect,
        southRect,
        eastRect;

    halo = TP.byId('SherpaHalo', TP.sys.getUIRoot());

    ourRect = this.getOffsetParent().getGlobalRect(false);

    targetRect = aTPElement.getHaloRect(halo);
    targetRect.subtractByPoint(ourRect.getXYPoint());

    x = targetRect.getX();
    y = targetRect.getY();

    northRect = TP.rtc(x, y, targetRect.getWidth(), 0.0);
    this.get('northguide').setOffsetPositionAndSize(northRect);

    westRect = TP.rtc(x, y, 0.0, targetRect.getHeight());
    this.get('westguide').setOffsetPositionAndSize(westRect);

    southRect = TP.rtc(x, y + targetRect.getHeight(), targetRect.getWidth(), 0.0);
    this.get('southguide').setOffsetPositionAndSize(southRect);

    eastRect = TP.rtc(x + targetRect.getWidth(), y, 0.0, targetRect.getHeight());
    this.get('eastguide').setOffsetPositionAndSize(eastRect);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineHandler('CanvasChanged',
function(aSignal) {

    /**
     * @method handleCanvasChanged
     * @summary Handles notifications of the canvas changing from the Sherpa
     *     object.
     * @param {TP.sig.CanvasChanged} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    var halo,
        haloTarget;

    if (this.get('$isRecasting')) {
        return this;
    }

    halo = TP.byId('SherpaHalo', this.getNativeDocument());
    haloTarget = halo.get('currentTargetTPElem');

    this.resizeToTarget(haloTarget);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineHandler('DOMDragMove',
function(aSignal) {

    /**
     * @method handleDOMDragMove
     * @summary Handles notification of when the receiver might be...
     * @param {TP.sig.DOMDragMove} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    var modifyingRule,

        halo,
        haloTarget,

        targetElem,

        currentTargetBox,

        targetWidth,
        targetHeight;

    modifyingRule = this.get('$currentModifyingRule');
    if (TP.notValid(modifyingRule)) {
        return this;
    }

    halo = TP.byId('SherpaHalo', this.getNativeDocument());
    haloTarget = halo.get('currentTargetTPElem');

    targetElem = haloTarget.getNativeNode();

    currentTargetBox = TP.elementGetBorderBox(targetElem);

    targetWidth = aSignal.getPageXAdjustedFor(targetElem) -
                    currentTargetBox.at('left');
    targetHeight = aSignal.getPageYAdjustedFor(targetElem) -
                    currentTargetBox.at('top');

    TP.styleRuleSetProperty(modifyingRule, 'width', targetWidth + 'px', false);
    TP.styleRuleSetProperty(modifyingRule, 'height', targetHeight + 'px', false);

    halo.moveAndSizeToTarget();

    this.resizeToTarget(haloTarget);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineHandler('DOMDragUp',
function(aSignal) {

    /**
     * @method handleDOMDragUp
     * @summary Handles notification of when the receiver might be...
     * @param {TP.sig.DOMDragUp} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    var modifyingRule,

        halo,
        haloTarget,

        targetElem,

        currentTargetBox,

        targetWidth,
        targetHeight;

    modifyingRule = this.get('$currentModifyingRule');
    if (TP.notValid(modifyingRule)) {
        return this;
    }

    halo = TP.byId('SherpaHalo', this.getNativeDocument());
    haloTarget = halo.get('currentTargetTPElem');

    targetElem = haloTarget.getNativeNode();

    currentTargetBox = TP.elementGetBorderBox(targetElem);

    targetWidth = aSignal.getPageXAdjustedFor(targetElem) -
                    currentTargetBox.at('left');
    targetHeight = aSignal.getPageYAdjustedFor(targetElem) -
                    currentTargetBox.at('top');

    TP.styleRuleSetProperty(modifyingRule, 'width', targetWidth + 'px', true);
    TP.styleRuleSetProperty(modifyingRule, 'height', targetHeight + 'px', true);

    halo.moveAndSizeToTarget();

    this.set('$currentModifyingRule', null);

    this.setAttribute('hidden', true);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineHandler('PclassClosedChange',
function(aSignal) {

    /**
     * @method handlePclassClosedChangeFromSherpaHUD
     * @summary Handles notifications of HUD closed change signals.
     * @param {TP.sig.PclassClosedChange} aSignal The TIBET signal which
     *     triggered this method.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    return this;
}, {
    origin: 'SherpaHUD'
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineHandler('HaloDidBlur',
function(aSignal) {

    /**
     * @method handleHaloDidBlur
     * @summary Handles notifications of when the halo blurs on an object.
     * @param {TP.sig.HaloDidBlur} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    var sherpaInst;

    sherpaInst = TP.bySystemId('Sherpa');
    this.ignore(sherpaInst, 'CanvasChanged');

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineHandler('HaloDidFocus',
function(aSignal) {

    /**
     * @method handleHaloDidFocus
     * @summary Handles notifications of when the halo focuses on an object.
     * @param {TP.sig.HaloDidFocus} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    var haloTarget,

        sherpaInst;

    haloTarget = aSignal.at('haloTarget');

    this.resizeToTarget(haloTarget);

    sherpaInst = TP.bySystemId('Sherpa');
    this.observe(sherpaInst, 'CanvasChanged');

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineMethod('setAttrHidden',
function(beHidden) {

    /**
     * @method setAttrHidden
     * @summary Sets the 'hidden' attribute of the receiver. This causes the
     *     halo to show or hide itself independent of whether it's focused or
     *     not.
     * @param {Boolean} beHidden Whether or not the halo should be hidden.
     * @returns {Boolean} Whether the receiver's state is hidden.
     */

    var wasHidden,

        toolsLayer,

        haloTPElem,
        targetTPElem,

        stylesHUD,
        modifyingRule;

    wasHidden = TP.bc(this.getAttribute('hidden'));

    if (wasHidden === beHidden) {
        //  Exit here - no need to call up to our supertype to toggle the
        //  attribute, since it already has the value we desire.
        return wasHidden;
    }

    //  Need to 'call up' to make sure the attribute value is actually captured.
    this.callNextMethod();

    toolsLayer = TP.bySystemId('Sherpa').getToolsLayer();

    haloTPElem = TP.byId('SherpaHalo', this.getNativeDocument());

    targetTPElem = haloTPElem.get('currentTargetTPElem');

    //  If we don't have a valid target element, then we exit here *without
    //  calling up to our supertype*. We don't want the attribute to be
    //  toggled, but we don't have valid target to show the halo on.
    if (TP.notValid(targetTPElem)) {
        return this;
    }

    if (TP.isTrue(beHidden)) {

        this.ignore(TP.core.Mouse, 'TP.sig.DOMDragMove');
        this.ignore(TP.core.Mouse, 'TP.sig.DOMDragUp');

        /*
        this.ignore(this, 'TP.sig.DOMMouseMove');
        this.ignore(this, 'TP.sig.DOMMouseOver');
        this.ignore(this, 'TP.sig.DOMMouseOut');

        this.ignore(TP.core.Mouse, 'TP.sig.DOMMouseWheel');

        this.ignore(TP.sys.getUICanvas().getDocument(),
                    TP.ac('TP.sig.MutationAttach',
                            'TP.sig.MutationDetach',
                            'TP.sig.MutationStyleChange'));

        this.ignore(this.getDocument(),
                    TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));
        this.ignore(TP.sys.getUICanvas().getDocument(),
                    TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));

        this.ignore(TP.bySystemId('Sherpa'), 'CanvasChanged');

        this.set('haloRect', null);
        */

        toolsLayer.removeAttribute('activetool');

    } else {

        this.observe(TP.core.Mouse, 'TP.sig.DOMDragMove');
        this.observe(TP.core.Mouse, 'TP.sig.DOMDragUp');

        /*
        //  If we don't have a valid target element, then we exit here *without
        //  calling up to our supertype*. We don't want the attribute to be
        //  toggled, but we don't have valid target to show the halo on.
        if (TP.notValid(this.get('currentTargetTPElem'))) {
            return this;
        }

        this.observe(this, 'TP.sig.DOMMouseMove');
        this.observe(this, 'TP.sig.DOMMouseOver');
        this.observe(this, 'TP.sig.DOMMouseOut');

        this.observe(TP.core.Mouse, 'TP.sig.DOMMouseWheel');

        this.observe(TP.sys.getUICanvas().getDocument(),
                    TP.ac('TP.sig.MutationAttach',
                            'TP.sig.MutationDetach',
                            'TP.sig.MutationStyleChange'));

        this.observe(this.getDocument(),
                        TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));
        this.observe(TP.sys.getUICanvas().getDocument(),
                        TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));

        this.observe(TP.bySystemId('Sherpa'), 'CanvasChanged');
        */

        this.resizeToTarget(targetTPElem);

        stylesHUD = TP.byId('StylesHUD', TP.win('UIROOT'));
        modifyingRule = stylesHUD.getModifiableRule();

        if (TP.notValid(modifyingRule)) {
            return this;
        }

        this.set('$currentModifyingRule', modifyingRule);

        toolsLayer.setAttribute('activetool', this.getLocalID());
    }

    //  Need to 'call up' to make sure the attribute value is actually captured.
    return beHidden;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsManipulator.Inst.defineMethod('setup',
function() {

    /**
     * @method setup
     * @summary Perform the initial setup for the receiver.
     * @returns {TP.sherpa.dimensionsManipulator} The receiver.
     */

    /*
    this.observe(TP.byId('SherpaHalo', TP.sys.getUIRoot()),
                    TP.ac('TP.sig.HaloDidFocus', 'TP.sig.HaloDidBlur'));

    this.observe(TP.byId('SherpaHUD', this.getNativeDocument()),
                            'PclassClosedChange');

    this.observe(TP.ANY, TP.ac('NodeWillRecast', 'NodeDidRecast'));
    */

    return this;
});

//  ------------------------------------------------------------------------
//  end
//  ========================================================================
