//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 */
//  ========================================================================

/**
 * @type {TP.sherpa.dimensionsmanipulator}
 */

//  ------------------------------------------------------------------------

TP.sherpa.Element.defineSubtype('sherpa.dimensionsmanipulator');

//  ------------------------------------------------------------------------
//  Instance Attributes
//  ------------------------------------------------------------------------

TP.sherpa.dimensionsmanipulator.Inst.defineAttribute('leftManipulator',
    TP.cpc('> svg|svg > svg|svg#leftManipulator', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsmanipulator.Inst.defineAttribute('rightManipulator',
    TP.cpc('> svg|svg > svg|line#rightManipulator', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsmanipulator.Inst.defineAttribute('topManipulator',
    TP.cpc('> svg|svg > svg|line#topManipulator', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsmanipulator.Inst.defineAttribute('bottomManipulator',
    TP.cpc('> svg|svg > svg|line#bottomManipulator', TP.hc('shouldCollapse', true)));

TP.sherpa.dimensionsmanipulator.Inst.defineAttribute('$currentlyMovingManipulator');

//  ------------------------------------------------------------------------
//  Instance Methods
//  ------------------------------------------------------------------------

TP.sherpa.dimensionsmanipulator.Inst.defineMethod('moveAndSizeToTarget',
function(newTargetTPElem, shouldUnhide) {

    /**
     * @method moveAndSizeToTarget
     * @summary Moves and sizes the receiver to the supplied target element or
     *     to the current target if a new target element isn't supplied (the
     *     current target's visual location might have shifted).
     * @param {TP.dom.ElementNode|undefined} [newTargetTPElem] The new target
     *     to move and resize the receiver around.
     * @param {Boolean} [shouldUnhide=false] Whether or not the halo should be
     *     shown if it's hidden. Some scenarios allow the halo to be hidden, but
     *     still focused.
     * @returns {TP.sherpa.halo} The receiver.
     */

    var handler,

        haloTPElem,

        theRect,
        ourRect,

        x,
        y,
        width,
        height,

        offset,

        grabberSize,
        halfGrabberSize,
        manWidth,
        manHeight,

        topManipulator,
        bottomManipulator,
        leftManipulator,
        rightManipulator;

    if (TP.notValid(newTargetTPElem)) {
        return this;
    }

    //  If the target element isn't ready to render (maybe its stylesheet hasn't
    //  loaded yet), then set up a handler to wait for the element to signal a
    //  TP.sig.DOMReady signal and then invoke this method again.
    if (!newTargetTPElem.isReadyToRender()) {

        handler = function() {
            handler.ignore(newTargetTPElem, 'TP.sig.DOMReady');
            this.moveAndSizeToTarget(newTargetTPElem, shouldUnhide);
        }.bind(this);

        handler.observe(newTargetTPElem, 'TP.sig.DOMReady');

        return this;
    } else {
        newTargetTPElem.render();
    }

    haloTPElem = TP.byId('SherpaHalo', this.getNativeDocument());

    //  Grab rect for the new target. Note that this will be supplied in
    //  *global* coordinates.
    theRect = newTargetTPElem.getHaloRect(haloTPElem);

    //  Given that the halo rect returned by one of the targets above is in
    //  *global* coordinates, we need to sure to adjust for the fact that
    //  the halo itself might be in a container that's positioned. We go to
    //  the halo's offset parent and get it's global rectangle.

    //  Note that we do this because the halo is currently hidden and won't
    //  give proper coordinates when computing relative to its offset
    //  parent.
    ourRect = haloTPElem.getOffsetParent().getGlobalRect(false);

    theRect.subtractByPoint(ourRect.getXYPoint());

    x = theRect.getX();
    y = theRect.getY();
    width = theRect.getWidth();
    height = theRect.getHeight();

    offset = 0;

    grabberSize = 20;
    halfGrabberSize = grabberSize / 2;

    /*
    topManipulator = this.get('topManipulator');
    topManipulator.setAttribute('x1', x + offset);
    topManipulator.setAttribute('y1', y);
    topManipulator.setAttribute('x2', x + width - offset);
    topManipulator.setAttribute('y2', y);

    bottomManipulator = this.get('bottomManipulator');
    bottomManipulator.setAttribute('x1', x + offset);
    bottomManipulator.setAttribute('y1', y + height);
    bottomManipulator.setAttribute('x2', x + width - offset);
    bottomManipulator.setAttribute('y2', y + height);
    */

    manWidth = grabberSize;
    manHeight = height + (grabberSize * 2);

    leftManipulator = this.get('leftManipulator');
    //leftManipulator.setAttribute('x', x - halfGrabberSize);
    //leftManipulator.setAttribute('y', y - grabberSize);
    x -= 12;

    leftManipulator.setAttribute('x', x);
    leftManipulator.setAttribute('y', y);
    leftManipulator.setAttribute('width', manWidth);
    leftManipulator.setAttribute('height', manHeight);
    leftManipulator.setAttribute('viewBox', '0 0 ' + manWidth + ' ' + manHeight);

    /*
    var line = leftManipulator.get(TP.cpc(' svg|line')).first();
    line.setAttribute('y2', (grabberSize - offset) + height);

    var endGrabber = leftManipulator.get(TP.cpc(' svg|use')).at(1);
    endGrabber.setAttribute('y', (grabberSize - (offset * 2)) + height);
    */

    /*
    rightManipulator = this.get('rightManipulator');
    rightManipulator.setAttribute('x1', x + width);
    rightManipulator.setAttribute('y1', y + offset);
    rightManipulator.setAttribute('x2', x + width);
    rightManipulator.setAttribute('y2', y + height - offset);
    */

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsmanipulator.Inst.defineMethod('setAttrHidden',
function(beHidden) {

    /**
     * @method setAttrHidden
     * @summary Sets the 'hidden' attribute of the receiver. This causes the
     *     halo to show or hide itself independent of whether it's focused or
     *     not.
     * @param {Boolean} beHidden Whether or not the halo should be hidden.
     * @returns {Boolean} Whether the receiver's state is hidden.
     */

    var wasHidden,

        toolsLayer,

        haloTPElem,
        targetTPElem;

    wasHidden = TP.bc(this.getAttribute('hidden'));

    if (wasHidden === beHidden) {
        //  Exit here - no need to call up to our supertype to toggle the
        //  attribute, since it already has the value we desire.
        return this;
    }

    toolsLayer = TP.bySystemId('Sherpa').getToolsLayer();

    haloTPElem = TP.byId('SherpaHalo', this.getNativeDocument());

    if (TP.isTrue(beHidden)) {
        /*
        this.ignore(this, 'TP.sig.DOMMouseMove');
        this.ignore(this, 'TP.sig.DOMMouseOver');
        this.ignore(this, 'TP.sig.DOMMouseOut');

        this.ignore(TP.core.Mouse, 'TP.sig.DOMMouseWheel');

        this.ignore(TP.sys.getUICanvas().getDocument(),
                    TP.ac('TP.sig.MutationAttach',
                            'TP.sig.MutationDetach',
                            'TP.sig.MutationStyleChange'));

        this.ignore(this.getDocument(),
                    TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));
        this.ignore(TP.sys.getUICanvas().getDocument(),
                    TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));

        this.ignore(TP.bySystemId('Sherpa'), 'CanvasChanged');
        */

        toolsLayer.removeAttribute('activetool');

    } else {

        targetTPElem = haloTPElem.get('currentTargetTPElem');

        //  If we don't have a valid target element, then we exit here *without
        //  calling up to our supertype*. We don't want the attribute to be
        //  toggled, but we don't have valid target to show the halo on.
        if (TP.notValid(targetTPElem)) {
            return this;
        }

        /*
        this.observe(this, 'TP.sig.DOMMouseMove');
        this.observe(this, 'TP.sig.DOMMouseOver');
        this.observe(this, 'TP.sig.DOMMouseOut');

        this.observe(TP.core.Mouse, 'TP.sig.DOMMouseWheel');

        this.observe(TP.sys.getUICanvas().getDocument(),
                    TP.ac('TP.sig.MutationAttach',
                            'TP.sig.MutationDetach',
                            'TP.sig.MutationStyleChange'));

        this.observe(this.getDocument(),
                        TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));
        this.observe(TP.sys.getUICanvas().getDocument(),
                        TP.ac('TP.sig.DOMResize', 'TP.sig.DOMScroll'));

        this.observe(TP.bySystemId('Sherpa'), 'CanvasChanged');
        */

        this.moveAndSizeToTarget(targetTPElem);

        toolsLayer.setAttribute('activetool', this.getLocalID());
    }

    //  Need to 'call up' to make sure the attribute value is actually captured.
    return this.callNextMethod();
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsmanipulator.Inst.defineHandler('TP.sig.DOMDragMove',
function(aSignal) {

    /**
     * @method handleTP.sig.DOMDragMove
     * @summary
     * @param {TP.sig.DOMDragMove} aSignal The TIBET signal which triggered this
     *     method.
     * @returns {TP.sherpa.dimensionsmanipulator} The receiver.
     */

    var newX,
        leftManipulator;

    newX = aSignal.get('offsetX') - 9;

    leftManipulator = this.get('$currentlyMovingManipulator');
    leftManipulator.setAttribute('x', newX);

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsmanipulator.Inst.defineHandler('StartMoveLeft',
function(aSignal) {

    /**
     * @method handleStartMoveLeft
     * @summary
     * @param {TP.sig.StartMoveLeft} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsmanipulator} The receiver.
     */

    this.set('$currentlyMovingManipulator', this.get('leftManipulator'));

    this.setAttribute('movingleft', true);

    this.observe(TP.core.Mouse, TP.ac('TP.sig.DOMDragMove', 'TP.sig.DOMDragUp'));

    return this;
});

//  ------------------------------------------------------------------------

TP.sherpa.dimensionsmanipulator.Inst.defineHandler('TP.sig.DOMDragUp',
function(aSignal) {

    /**
     * @method TP.sig.DOMDragUp
     * @summary
     * @param {TP.sig.StartMoveLeft} aSignal The TIBET signal which triggered
     *     this method.
     * @returns {TP.sherpa.dimensionsmanipulator} The receiver.
     */

    this.observe(TP.core.Mouse, TP.ac('TP.sig.DOMDragMove', 'TP.sig.DOMDragUp'));

    this.removeAttribute('movingleft');

    this.set('$currentlyMovingManipulator', null);

    return this;
});

//  ------------------------------------------------------------------------
//  end
//  ========================================================================
